__author__ = "xlinfr"

import numpy as np
try:
    from osgeo import gdal, osr
    from osgeo.gdalconst import GDT_Float32
except:
    pass

# Slope and aspect used in SEBE and Wall aspect
def get_ders(dsm, scale):
    # dem,_,_=read_dem_grid(dem_file)
    dx = 1 / scale
    # dx=0.5
    fy, fx = np.gradient(dsm, dx, dx)
    asp, grad = cart2pol(fy, fx, "rad")
    slope = np.arctan(grad)
    asp = asp * -1
    asp = asp + (asp < 0) * (np.pi * 2)
    return slope, asp


def cart2pol(x, y, units="deg"):
    radius = np.sqrt(x**2 + y**2)
    theta = np.arctan2(y, x)
    if units in ["deg", "degs"]:
        theta = theta * 180 / np.pi
    return theta, radius

def saveraster(gdal_data, filename, raster):
    rows = gdal_data.RasterYSize
    cols = gdal_data.RasterXSize

    outDs = gdal.GetDriverByName("GTiff").Create(filename, cols, rows, int(1), GDT_Float32)
    outBand = outDs.GetRasterBand(1)

    # write the data
    outBand.WriteArray(raster, 0, 0)
    # flush data to disk, set the NoData value and calculate stats
    outBand.FlushCache()
    outBand.SetNoDataValue(-9999)

    # georeference the image and set the projection
    outDs.SetGeoTransform(gdal_data.GetGeoTransform())
    outDs.SetProjection(gdal_data.GetProjection())

def saverasternd(gdal_data, filename, raster):
    rows = gdal_data.RasterYSize
    cols = gdal_data.RasterXSize

    outDs = gdal.GetDriverByName("GTiff").Create(filename, cols, rows, int(1), GDT_Float32)
    outBand = outDs.GetRasterBand(1)

    # write the data
    outBand.WriteArray(raster, 0, 0)
    # flush data to disk, set the NoData value and calculate stats
    outBand.FlushCache()
    # outBand.SetNoDataValue(-9999)

    # georeference the image and set the projection
    outDs.SetGeoTransform(gdal_data.GetGeoTransform())
    outDs.SetProjection(gdal_data.GetProjection())

def xy2latlon(crsWtkIn, x, y):
    old_cs = osr.SpatialReference()
    old_cs.ImportFromWkt(crsWtkIn)

    wgs84_wkt = """
    GEOGCS["WGS 84",
        DATUM["WGS_1984",
            SPHEROID["WGS 84",6378137,298.257223563,
                AUTHORITY["EPSG","7030"]],
            AUTHORITY["EPSG","6326"]],
        PRIMEM["Greenwich",0,
            AUTHORITY["EPSG","8901"]],
        UNIT["degree",0.01745329251994328,
            AUTHORITY["EPSG","9122"]],
        AUTHORITY["EPSG","4326"]]"""

    new_cs = osr.SpatialReference()
    new_cs.ImportFromWkt(wgs84_wkt)

    transform = osr.CoordinateTransformation(old_cs, new_cs)

    lonlat = transform.TransformPoint(x, y)

    gdalver = float(gdal.__version__[0])
    if gdalver >= 3.:
        lon = lonlat[1] #changed to gdal 3
        lat = lonlat[0] #changed to gdal 3
    else:
        lon = lonlat[0] #changed to gdal 2
        lat = lonlat[1] #changed to gdal 2

    return lat, lon

def xy2latlon_fromraster(crsWtkIn, gdal_dsm):
    # Get latlon from grid coordinate system
    old_cs = osr.SpatialReference()
    # dsm_ref = dsmlayer.crs().toWkt()
    old_cs.ImportFromWkt(crsWtkIn)

    wgs84_wkt = """
    GEOGCS["WGS 84",
        DATUM["WGS_1984",
            SPHEROID["WGS 84",6378137,298.257223563,
                AUTHORITY["EPSG","7030"]],
            AUTHORITY["EPSG","6326"]],
        PRIMEM["Greenwich",0,
            AUTHORITY["EPSG","8901"]],
        UNIT["degree",0.01745329251994328,
            AUTHORITY["EPSG","9122"]],
        AUTHORITY["EPSG","4326"]]"""

    new_cs = osr.SpatialReference()
    new_cs.ImportFromWkt(wgs84_wkt)

    transform = osr.CoordinateTransformation(old_cs, new_cs)
    widthx = gdal_dsm.RasterXSize
    heightx = gdal_dsm.RasterYSize
    geotransform = gdal_dsm.GetGeoTransform()
    minx = geotransform[0]
    miny = geotransform[3] + widthx * geotransform[4] + heightx * geotransform[5]
    
    lonlat = transform.TransformPoint(minx, miny)
    gdalver = float(gdal.__version__[0])
    if gdalver == 3.:
        lon = lonlat[1] #changed to gdal 3
        lat = lonlat[0] #changed to gdal 3
    else:
        lon = lonlat[0] #changed to gdal 2
        lat = lonlat[1] #changed to gdal 2
    scale = 1 / geotransform[1]

    return lat, lon, scale, minx, miny


from typing import Any, Tuple

from ...class_configs import SolweigConfig
from .solweig_runner import SolweigRun
from .wallOfInterest import pointOfInterest


class SolweigRunQgis(SolweigRun):
    """Run SOLWEIG in QGIS environment."""

    def __init__(self, config_path_str: str, params_json_path: str, feedback: Any):
        """ """
        config = SolweigConfig()
        config.from_file(config_path_str)
        super().__init__(config, params_json_path)
        self.progress = feedback

    def prep_progress(self, num: int) -> None:
        """Prepare progress for environment."""
        self.iters_total = num
        self.iters_count = 0

    def iter_progress(self) -> bool:
        """Iterate progress."""
        self.progress.setProgress(int(self.iters_count * (100.0 / self.iters_total)))  # move progressbar forward
        if self.progress.isCanceled():
            self.progress.setProgressText("Calculation cancelled")
            return False
        return True

    def load_poi_data(self) -> Tuple[Any, Any]:
        """Load points of interest (POIs) from a file."""
        scale = 1 / self.raster_data.trf_arr[1]
        poi_names, poi_pixel_xys = pointOfInterest(
            self.config.poi_file, self.config.poi_field, scale, self.raster_data.trf_arr
        )
        self.poi_names = poi_names
        self.poi_pixel_xys = poi_pixel_xys

    def save_poi_results(self) -> None:
        """Save points of interest (POIs) results to a text file with geographic coordinates."""
        if not self.poi_results:
            return
        # Extract header from first result and add east/northing
        header = ["name"] + list(self.poi_results[0].keys()) + ["easting", "northing"]
        # Write results to file
        output_path = self.config.output_dir + "/POI_results.txt"
        with open(output_path, "w") as f:
            f.write("\t".join(header) + "\n")
            for result in self.poi_pixel_xys:
                lng = result["col_idx"] * self.raster_data.trf_arr[1] + self.raster_data.trf_arr[0]
                lat = result["row_idx"] * self.raster_data.trf_arr[1] + self.raster_data.trf_arr[3]
                row_values = list(result.values()) + [lng, lat]
                f.write("\t".join(map(str, row_values)) + "\n")

    def load_woi_data(self) -> Tuple[Any, Any]:
        """Load walls of interest (WOIs) from a file."""
        scale = 1 / self.raster_data.trf_arr[1]
        woi_names, woi_pixel_xys = pointOfInterest(
            self.config.woi_file, self.config.woi_field, scale, self.raster_data.trf_arr
        )
        self.woi_names = woi_names
        self.woi_pixel_xys = woi_pixel_xys

    def save_woi_results(self) -> None:
        """Save walls of interest (WOIs) results to a text file with geographic coordinates."""
        if not self.woi_results:
            return
        # Extract header from first result and add lng/lat
        header = ["name"] + list(self.woi_results[0].keys()) + ["lng", "lat"]
        # Write results to file
        output_path = self.config.output_dir + "/WOI_results.txt"
        with open(output_path, "w") as f:
            f.write("\t".join(header) + "\n")
            for result in self.woi_pixel_xys:
                lng = result["col_idx"] * self.raster_data.trf_arr[1] + self.raster_data.trf_arr[0]
                lat = result["row_idx"] * self.raster_data.trf_arr[1] + self.raster_data.trf_arr[3]
                row_values = list(result.values()) + [lng, lat]
                f.write("\t".join(map(str, row_values)) + "\n")



# -*- coding: utf-8 -*-
from __future__ import division
import numpy as np
from math import radians
# from scipy.ndimage.filters import median_filter


def shadowingfunction_wallheight_13(a, azimuth, altitude, scale, walls, aspect):
    """
    This m.file calculates shadows on a DSM and shadow height on building
    walls.
    
    INPUTS:
    a = DSM
    azimuth and altitude = sun position
    scale= scale of DSM (1 meter pixels=1, 2 meter pixels=0.5)
    walls= pixel row 'outside' buildings. will be calculated if empty
    aspect = normal aspect of buildings walls
    
    OUTPUT:
    sh=ground and roof shadow
    wallsh = height of wall that is in shadow
    wallsun = hieght of wall that is in sun
    
    Fredrik Lindberg 2012-03-19
    fredrikl@gvc.gu.se
    
     Utdate 2013-03-13 - bugfix for walls alinged with sun azimuths

    :param a:
    :param azimuth:
    :param altitude:
    :param scale:
    :param walls:
    :param aspect:
    :return:
    """

    if not walls.size:
        """
        walls = ordfilt2(a,4,[0 1 0; 1 0 1; 0 1 0])
        walls = walls-a
        walls[walls < 3]=0
        sizex = np.shape(a)[0]    #might be wrong
        sizey = np.shape(a)[1]
        dirwalls = filter1Goodwin_as_aspect_v3(walls,sizex,sizey,scale,a);
        aspect = dirwalls*np.pi/180
        """

    # conversion
    # degrees = np.pi/180
    azimuth = radians(azimuth)
    altitude = radians(altitude)

    # measure the size of the image
    sizex = np.shape(a)[0]
    sizey = np.shape(a)[1]

    # initialise parameters
    f = np.copy(a)
    dx = 0
    dy = 0
    dz = 0
    temp = np.zeros((sizex, sizey))
    wallbol = (walls > 0).astype(float)
    
    # other loop parameters
    amaxvalue = np.max(a)
    pibyfour = np.pi/4
    threetimespibyfour = 3 * pibyfour
    fivetimespibyfour = 5 * pibyfour
    seventimespibyfour = 7 * pibyfour
    sinazimuth = np.sin(azimuth)
    cosazimuth = np.cos(azimuth)
    tanazimuth = np.tan(azimuth)
    signsinazimuth = np.sign(sinazimuth)
    signcosazimuth = np.sign(cosazimuth)
    dssin = np.abs(1/sinazimuth)
    dscos = np.abs(1/cosazimuth)
    tanaltitudebyscale = np.tan(altitude)/scale

    index = 1

    # main loop
    while (amaxvalue >= dz) and (np.abs(dx) < sizex) and (np.abs(dy) < sizey):

        if (pibyfour <= azimuth and azimuth < threetimespibyfour) or \
                (fivetimespibyfour <= azimuth and azimuth < seventimespibyfour):
            dy = signsinazimuth * index
            dx = -1 * signcosazimuth * np.abs(np.round(index / tanazimuth))
            ds = dssin
        else:
            dy = signsinazimuth * np.abs(np.round(index * tanazimuth))
            dx = -1 * signcosazimuth * index
            ds = dscos

        # note: dx and dy represent absolute values while ds is an incremental value
        dz = ds * index * tanaltitudebyscale
        temp[0:sizex, 0:sizey] = 0

        absdx = np.abs(dx)
        absdy = np.abs(dy)

        xc1 = int((dx+absdx)/2)
        xc2 = int(sizex+(dx-absdx)/2)
        yc1 = int((dy+absdy)/2)
        yc2 = int(sizey+(dy-absdy)/2)

        xp1 = int(-((dx-absdx)/2))
        xp2 = int(sizex-(dx+absdx)/2)
        yp1 = int(-((dy-absdy)/2))
        yp2 = int(sizey-(dy+absdy)/2)

        temp[xp1:xp2, yp1:yp2] = a[xc1:xc2, yc1:yc2] - dz
        f = np.fmax(f, temp) #Moving building shadow

        index = index + 1

    # Removing walls in shadow due to selfshadowing
    azilow = azimuth-np.pi/2
    azihigh = azimuth+np.pi/2

    if azilow >= 0 and azihigh < 2*np.pi:    # 90 to 270  (SHADOW)
        facesh = (np.logical_or(aspect < azilow, aspect >= azihigh).astype(float)-wallbol+1)
    elif azilow < 0 and azihigh <= 2*np.pi:    # 0 to 90
        azilow = azilow + 2*np.pi
        facesh = np.logical_or(aspect > azilow, aspect <= azihigh) * -1 + 1    # (SHADOW)    # check for the -1
    elif azilow > 0 and azihigh >= 2*np.pi:    # 270 to 360
        azihigh = azihigh-2*np.pi
        facesh = np.logical_or(aspect > azilow, aspect <= azihigh)*-1 + 1    # (SHADOW)

    sh = np.copy(f-a)    # shadow volume
    facesun = np.logical_and(facesh + (walls > 0).astype(float) == 1, walls > 0).astype(float)
    wallsun = np.copy(walls-sh)
    wallsun[wallsun < 0] = 0
    wallsun[facesh == 1] = 0    # Removing walls in "self"-shadow
    wallsh = np.copy(walls-wallsun)

    sh = np.logical_not(np.logical_not(sh)).astype(float)
    sh = sh * -1 + 1

    return sh, wallsh, wallsun, facesh, facesun



import numpy as np
import pandas as pd

from copy import deepcopy

from . import emissivity_models
from . import patch_radiation
from . import sunlit_shaded_patches
from . import patch_radiation

def anisotropic_sky(shmat, vegshmat, vbshvegshmat, solar_altitude, solar_azimuth, asvf, cyl, 
                    esky, L_patches, wallScheme, voxelTable, voxelMaps, steradians, Ta, Tgwall, ewall, Lup,
                    radI, radD, radG, lv, albedo, anisotropic_diffuse, diffsh, shadow, KupE, KupS, KupW, KupN, current_step):
    
    '''This function estimates short- and longwave radiation from an anisotropic sky....'''
    # Stefan-Boltzmann's Constant
    SBC = 5.67051e-8

    # Degrees to radians
    deg2rad = np.pi / 180

    # Shape of rasters
    rows = shmat.shape[0]
    cols = shmat.shape[1]

    # Define shortwave variables
    KsideD = np.zeros((rows, cols))
    KsideI = np.zeros((rows, cols))
    Kside = np.zeros((rows, cols))
    Kref_sun = np.zeros((rows, cols))
    Kref_sh = np.zeros((rows, cols))
    Kref_veg = np.zeros((rows, cols))

    Keast = np.zeros((rows, cols))
    Kwest = np.zeros((rows, cols))
    Knorth = np.zeros((rows, cols))
    Ksouth = np.zeros((rows, cols))

    # Define longwave variables
    Ldown_sky = np.zeros((rows, cols))
    Ldown_veg = np.zeros((rows, cols))
    Ldown_sun = np.zeros((rows, cols))
    Ldown_sh = np.zeros((rows, cols))
    Ldown_ref = np.zeros((rows, cols))

    Lside_sky = np.zeros((rows, cols))
    Lside_veg = np.zeros((rows, cols))
    Lside_sun = np.zeros((rows, cols))
    Lside_sh = np.zeros((rows, cols))
    Lside_ref = np.zeros((rows, cols))

    Least = np.zeros((rows, cols))
    Lwest = np.zeros((rows, cols))
    Lnorth = np.zeros((rows, cols))
    Lsouth = np.zeros((rows, cols))

    # Unique altitudes for patches
    skyalt, skyalt_c = np.unique(L_patches[:, 0], return_counts=True)

    # Patch altitudes and azimuths
    patch_altitude = L_patches[:, 0]
    patch_azimuth = L_patches[:, 1]

    # Longwave radiation
    # Current model for anisotropic sky emissivity
    emis_m = 2

    # Unsworth & Monteith (1975)
    if emis_m == 1:
        patch_emissivity_normalized, esky_band = emissivity_models.model1(L_patches, esky, Ta)
    # Martin & Berdahl (1984)
    elif emis_m == 2:
        patch_emissivity_normalized, esky_band = emissivity_models.model2(L_patches, esky, Ta)
    # Bliss (1961)
    elif emis_m == 3:
        patch_emissivity_normalized, esky_band = emissivity_models.model3(L_patches, esky, Ta)

    # True = anisotropic sky, False = isotropic sky
    anisotropic_sky_ = True

    # Longwave based on spectral flux density (divide by pi)
    Ldown = np.zeros((patch_altitude.shape[0]))
    Lside = np.zeros((patch_altitude.shape[0]))
    Lnormal = np.zeros((patch_altitude.shape[0]))
    for temp_altitude in skyalt:
        # Anisotropic sky
        if anisotropic_sky_:
            temp_emissivity = esky_band[skyalt == temp_altitude]
        # Isotropic sky but with patches (need to switch anisotropic_sky to False)
        else:
            temp_emissivity = esky
        # Estimate longwave radiation on a horizontal surface (Ldown), vertical surface (Lside) and perpendicular (Lnormal)
        Ldown[patch_altitude == temp_altitude] = ((temp_emissivity * SBC * ((Ta + 273.15) ** 4)) / np.pi) * steradians[patch_altitude == temp_altitude] * np.sin(temp_altitude * deg2rad)
        Lside[patch_altitude == temp_altitude] = ((temp_emissivity * SBC * ((Ta + 273.15) ** 4)) / np.pi) * steradians[patch_altitude == temp_altitude] * np.cos(temp_altitude * deg2rad)
        Lnormal[patch_altitude == temp_altitude] = ((temp_emissivity * SBC * ((Ta + 273.15) ** 4)) / np.pi) * steradians[patch_altitude == temp_altitude]

    Lsky_normal = deepcopy(L_patches)
    Lsky_down = deepcopy(L_patches)
    Lsky_side = deepcopy(L_patches)

    Lsky_normal[:,2] = Lnormal
    Lsky_down[:,2] = Ldown
    Lsky_side[:,2] = Lside

    # Shortwave radiation
    if solar_altitude > 0:
        # Patch luminance
        patch_luminance = lv[:, 2]
        # Variable to fill with total sky diffuse shortwave radiation
        radTot = np.zeros(1)
        # Radiance fraction normalization
        for i in np.arange(patch_altitude.shape[0]):
            radTot += (patch_luminance[i] * steradians[i] * np.sin(patch_altitude[i] * deg2rad)) # Radiance fraction normalization
        lumChi = (patch_luminance * radD) / radTot

    for i in np.arange(patch_altitude.shape[0]):
        # Calculations for patches on sky, shmat = 1 = sky is visible
        temp_sky = ((shmat[:, :, i] == 1) & (vegshmat[:, :, i] == 1))
        
        # Calculations for patches that are vegetation, vegshmat = 0 = shade from vegetation
        temp_vegsh = ((vegshmat[:, :, i] == 0) | (vbshvegshmat[:, :, i] == 0))

        # Calculations for patches that are buildings, shmat = 0 = shade from buildings
        temp_vbsh = (1 - shmat[:, :, i]) * vbshvegshmat[:, :, i]
        temp_sh = (temp_vbsh == 1)
        if wallScheme == 1:
            # temp_vbsh = (voxelMaps[:, :, i] > 0) * vbshvegshmat[:, :, i]
            # temp_sh_w = (temp_vbsh == 1) * voxelMaps[:, :, i]
            temp_sh_w = temp_sh * voxelMaps[:, :, i]
            temp_sh_roof = temp_sh * (voxelMaps[:, :, i] == 0)

        # Estimate sunlit and shaded patches
        sunlit_patches, shaded_patches = sunlit_shaded_patches.shaded_or_sunlit(solar_altitude, solar_azimuth, patch_altitude[i], patch_azimuth[i], asvf)

        if cyl == 1:
            # Angle of incidence, np.cos(0) because cylinder - always perpendicular
            angle_of_incidence = np.cos(patch_altitude[i] * deg2rad) * np.cos(0) # * np.sin(np.pi / 2) \
            # Angle of incidence to horizontal surface
            angle_of_incidence_h = np.sin(patch_altitude[i] * deg2rad)

            ### CALCULATIONS FOR LONGWAVE RADIATION ###
            # Longwave radiation from sky
            Lside_sky_temp, Ldown_sky_temp, Least_temp, Lsouth_temp, Lwest_temp, Lnorth_temp = patch_radiation.longwave_from_sky(temp_sky, Lsky_side[i, 2], Lsky_down[i, 2], patch_azimuth[i])          
            
            Lside_sky += Lside_sky_temp
            Ldown_sky += Ldown_sky_temp
            Least += Least_temp
            Lsouth += Lsouth_temp
            Lwest += Lwest_temp
            Lnorth += Lnorth_temp

            # Longwave radiation from vegetation
            Lside_veg_temp, Ldown_veg_temp, \
                Least_temp, Lsouth_temp, Lwest_temp, Lnorth_temp = patch_radiation.longwave_from_veg(temp_vegsh, steradians[i], angle_of_incidence, angle_of_incidence_h, 
                                                                                                    patch_altitude[i], patch_azimuth[i], ewall, Ta)                   
            
            Lside_veg += Lside_veg_temp
            Ldown_veg += Ldown_veg_temp
            Least += Least_temp
            Lsouth += Lsouth_temp
            Lwest += Lwest_temp
            Lnorth += Lnorth_temp            

            # Longwave radiation from buildings
            if wallScheme == 0:
                azimuth_difference = np.abs(solar_azimuth - patch_azimuth[i])

                Lside_sun_temp, Lside_sh_temp, \
                Ldown_sun_temp, Ldown_sh_temp, \
                Least_temp, Lsouth_temp, Lwest_temp, Lnorth_temp = patch_radiation.longwave_from_buildings(temp_sh, steradians[i], angle_of_incidence, angle_of_incidence_h, 
                                                                                                         patch_azimuth[i], sunlit_patches, shaded_patches, 
                                                                                                         azimuth_difference, solar_altitude, ewall, Ta, Tgwall)

            else:
                azimuth_difference = np.abs(solar_azimuth - patch_azimuth[i])
                # print('Building pixels = ' + str(temp_sh.sum()))
                # print('Building pixels wall scheme = ' + str(np.sum(temp_sh_w > 0)))
                Lside_sun_temp, Lside_sh_temp, \
                Ldown_sun_temp, Ldown_sh_temp, \
                Least_temp, Lsouth_temp, Lwest_temp, Lnorth_temp = patch_radiation.longwave_from_buildings_wallScheme(temp_sh_w, voxelTable, steradians[i], angle_of_incidence, angle_of_incidence_h, 
                                                                                                         patch_azimuth[i])                

                Lside_sun_r_temp, Lside_sh_r_temp, \
                Ldown_sun_r_temp, Ldown_sh_r_temp, \
                Least_r_temp, Lsouth_r_temp, Lwest_r_temp, Lnorth_r_temp = patch_radiation.longwave_from_buildings(temp_sh_roof, steradians[i], angle_of_incidence, angle_of_incidence_h, 
                                                                                                         patch_azimuth[i], sunlit_patches, shaded_patches, 
                                                                                                         azimuth_difference, solar_altitude, ewall, Ta, Tgwall)

                Lside_sun_temp += Lside_sun_r_temp
                Lside_sh_temp += Lside_sh_r_temp
                Ldown_sun_temp += Ldown_sun_r_temp
                Ldown_sh_temp += Ldown_sh_r_temp
                Least_temp += Least_r_temp
                Lsouth_temp += Lsouth_r_temp
                Lwest_temp += Lwest_r_temp
                Lnorth_temp += Lnorth_r_temp

            Lside_sun += Lside_sun_temp
            Lside_sh += Lside_sh_temp
            Ldown_sun += Ldown_sun_temp
            Ldown_sh += Ldown_sh_temp
            Least += Least_temp
            Lsouth += Lsouth_temp
            Lwest += Lwest_temp
            Lnorth += Lnorth_temp 

            ### CALCULATIONS FOR SHORTWAVE RADIATION ###
            if solar_altitude > 0:
                # Shortwave radiation from sky
                KsideD += temp_sky * lumChi[i] * angle_of_incidence * steradians[i]
                
                # Shortwave reflected on sunlit surfaces
                # sunlit_surface = ((albedo * radG) / np.pi)
                sunlit_surface = ((albedo * (radI * np.cos(solar_altitude * deg2rad)) + (radD * 0.5)) / np.pi)
                # Shortwave reflected on shaded surfaces and vegetation
                shaded_surface = ((albedo * radD * 0.5) / np.pi)
                
                # Shortwave radiation from vegetation
                Kref_veg += shaded_surface * temp_vegsh * steradians[i] * angle_of_incidence
                
                # Shortwave radiation from buildings
                Kref_sun += sunlit_surface * sunlit_patches * temp_sh * steradians[i] * angle_of_incidence
                Kref_sh += shaded_surface * shaded_patches * temp_sh * steradians[i] * angle_of_incidence

    # Calculate reflected longwave in each patch
    for idx in np.arange(patch_altitude.shape[0]):
        # Angle of incidence, np.cos(0) because cylinder - always perpendicular
        angle_of_incidence = np.cos(patch_altitude[idx] * deg2rad) * np.cos(0) # * np.sin(np.pi / 2) \

        # Angle of incidence to horizontal surface
        angle_of_incidence_h = np.sin(patch_altitude[idx] * deg2rad)

        # Patches with reflected surfaces
        temp_sh = ((shmat[:,:,idx] == 0) | (vegshmat[:,:,idx] == 0) | (vbshvegshmat[:,:,idx] == 0))
        
        Lside_ref_temp, Ldown_ref_temp, \
        Least_temp, Lsouth_temp, Lwest_temp, Lnorth_temp = patch_radiation.reflected_longwave(temp_sh, steradians[idx], angle_of_incidence, angle_of_incidence_h, patch_azimuth[idx], Ldown_sky, Lup, ewall)

        Lside_ref += Lside_ref_temp
        Ldown_ref += Ldown_ref_temp
        Least += Least_temp
        Lsouth += Lsouth_temp
        Lwest += Lwest_temp
        Lnorth += Lnorth_temp

    # Sum of all Lside components (sky, vegetation, sunlit and shaded buildings, reflected)
    Lside = Lside_sky + Lside_veg + Lside_sh + Lside_sun + Lside_ref

    # Sum of all Lside components (sky, vegetation, sunlit and shaded buildings, reflected)
    Ldown = Ldown_sky + Ldown_veg + Ldown_sh + Ldown_sun + Ldown_ref

    ### Direct radiation ###
    if cyl == 1: ### Kside with cylinder ###
        KsideI = shadow * radI * np.cos(solar_altitude * deg2rad)

    if solar_altitude > 0:

        Kside = KsideI + KsideD + Kref_sun + Kref_sh + Kref_veg

        Keast = (KupE * 0.5)
        Kwest = (KupW * 0.5)
        Knorth = (KupN * 0.5)
        Ksouth = (KupS * 0.5)

    return Ldown, Lside, Lside_sky, Lside_veg, Lside_sh, Lside_sun, Lside_ref, Least, Lwest, Lnorth, Lsouth, \
           Keast, Ksouth, Kwest, Knorth, KsideI, KsideD, Kside, steradians, skyalt

    